/*
 * pad.cc --
 *	Based on a sample Tk widget and includes code from
 *	Practical Programming in Tcl and Tk
 *      by Brent B. Welch
 *
 *  parts of this code are copyrighted by Xerox and Prentice Hall
 *  see the file copyright.welch
 */

/*
 * Include the standard Tk header files.
 */
#include "tk.h"
#include <string.h>
#include <stdlib.h>


/*
 * The Pad typedef is a data structure that defines one instance
 * of the widget.  There might be many instances in one interface.
 */
typedef struct
{
	Tk_Window tkwin;	/* The window for the widget */
	Display *display;	/* X's handle on the display */
	Tcl_Interp *interp;	/* The interpreter of the widget */
	/*
	 * Pad-specific attributes.
	 */
	Pixmap pixmap;
	XFontStruct *fontPtr;	/* Font info for the text */
	GC pixGC;		/* Graphics context for the text display */
	XColor *foreground;	/* Color for the text */
	int flags;		/* Flags defined below */
	Tk_3DBorder background;	/* Color info for border & background */
	int width;		/* in pixels */
	int height;		/* in pixels */

	#if CLOCK
	int borderWidth;	/* Size of 3-D border */
	int relief;		/* Style of 3-D border */
	Tk_3DBorder background;	/* Color info for border & background */
	XColor *foreground;	/* Color for the text */

	XColor *highlight;	/* Color info for the highlight */
	int highlightWidth;	/* Thickness of highlight rim */

	XFontStruct *fontPtr;	/* Font info for the text */
	char *format;		/* Format for the pad display */
	/*
	 * Graphic contexts and other support.
	 */
	GC highlightGC;		/* Graphics context for the highlight */
	GC pixGC;		/* Graphics context for the text display */
	Tk_TimerToken token;	/* Handle on periodic callbacks */
	char *pad;		/* Pointer to the pad string */
	int numChars;		/* in the text */
	int textWidth;		/* in pixels */
	int textHeight;		/* in pixels */
	int flags;		/* Flags defined below */
	#endif
} Pad;

/*
 * Flag bits.
 *	REDRAW_PENDING  - a call to PadDisplay has been queued already.
 */
#define REDRAW_PENDING	0x1

/*
 * These configuration specs are used to automatically process
 * arguments to the command that creates a Pad and to the
 * configure operation that changes a widget.  They define the
 * command line flag, its type, its resource name, its class name,
 * and its default value.  Tk_Offset is a macro that gets the
 * address of the structure member so the Pad struct can
 * be modified directly.
 */

static Tk_ConfigSpec configSpecs[] =
{
	{TK_CONFIG_BORDER, "-background", "background", "Background",
		"light blue", Tk_Offset(Pad, background), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_BORDER, "-background", "background", "Background",
		"white", Tk_Offset(Pad, background), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
		(char *) NULL, 0, 0},

	{TK_CONFIG_PIXELS, "-width", "width", "Width",
		"200", Tk_Offset(Pad, width), 0},
	{TK_CONFIG_PIXELS, "-height", "height", "Height",
		"200", Tk_Offset(Pad, height), 0},

	#if CLOCK
	{TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
		(char *) NULL, 0, 0},
	{TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
		"2", Tk_Offset(Pad, borderWidth), 0},
	{TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
		"ridge", Tk_Offset(Pad, relief), 0},
	#endif

	{TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
		"black", Tk_Offset(Pad, foreground), 0},
	{TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
		(char *) NULL, 0, 0},

	#if CLOCK
	{TK_CONFIG_COLOR, "-highlightcolor", "highlightColor", "HighlightColor",
		"red", Tk_Offset(Pad, highlight),
		TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-highlightcolor", "highlightColor", "HighlightColor",
		"black", Tk_Offset(Pad, highlight),
		TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_PIXELS, "-highlightthickness", "highlightThickness",
		"HighlightThickness",
		"2", Tk_Offset(Pad, highlightWidth),
		0},

	{TK_CONFIG_STRING, "-format", "format", "Format",
		"%H:%M:%S", Tk_Offset(Pad, format), 0},
	#endif
	{TK_CONFIG_FONT, "-font", "font", "Font",
		"*courier-medium-r-normal-*-18-*", Tk_Offset(Pad, fontPtr),
		0},

	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
		(char *) NULL, 0, 0}
};

/*
 * Forward declarations.  _ANSI_ARGS_ is a macro that checks to
 * see if we are being compiled with an ANSI C compiler before
 * it puts the parameter types into the declaration.
 */
static int		PadConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    Pad *padPtr, int argc, char **argv,
			    int flags));
static void		PadDestroy _ANSI_ARGS_((char* clientData));
//static void		PadDestroy _ANSI_ARGS_((ClientData clientData));
static void		PadDisplay _ANSI_ARGS_((ClientData clientData));
static void		PadEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		PadInstanceCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));

/*
 * PadCmd --
 *	This is the command procedure for the "mywidget" Tcl command.
 *	It creates an instance of the widget.  As a side effect, a new
 *	Tcl command is created that operates on the new instance.
 *	This returns TCL_OK or TCL_ERROR if the widget could not be
 *	created.
 */
int
PadCmd(ClientData clientData, Tcl_Interp* interp, int argc, char** argv)
    /*ClientData clientData;	/* Main window associated with
				 * interpreter. */
    /*Tcl_Interp *interp;		/* Current interpreter. */
    /*int argc;			/* Number of arguments. */
    /*char **argv;		/* Argument strings. */
{
	Tk_Window main = (Tk_Window) clientData;
	Pad *padPtr;
	Tk_Window tkwin;
    
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args:  should be \"",
			    argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}

	tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	Tk_SetClass(tkwin, "Pad");
    
	/*
	 * Allocate and initialize the widget record.
	 */

	padPtr = (Pad *) ckalloc(sizeof(Pad));
	padPtr->tkwin = tkwin;
	padPtr->display = Tk_Display(tkwin);
	padPtr->interp = interp;
	padPtr->pixmap = (Pixmap ) 0;
	padPtr->background = NULL;
	padPtr->foreground = NULL;
	padPtr->fontPtr = NULL;
	padPtr->pixGC = None;
	padPtr->flags = 0;
	#if CLOCK
	padPtr->borderWidth = 0;
	padPtr->highlightWidth = 0;
	padPtr->relief = TK_RELIEF_FLAT;
	padPtr->background = NULL;
	padPtr->foreground = NULL;
	padPtr->highlight = NULL;
	padPtr->fontPtr = NULL;
	padPtr->textGC = None;
	padPtr->highlightGC = None;
	padPtr->token = NULL;
	padPtr->pad = NULL;
	padPtr->numChars = 0;
	padPtr->textWidth = 0;
	padPtr->textHeight = 0;
	padPtr->flags = 0;
	#endif

	/*
	 * Register a handler for when the window is
	 * exposed or resized.
	 */
	Tk_CreateEventHandler(padPtr->tkwin,
		ExposureMask|StructureNotifyMask|FocusChangeMask,
		PadEventProc, (ClientData) padPtr);
	/*
	 * Create a Tcl command that operates on the widget.
	 */
	Tcl_CreateCommand(interp, Tk_PathName(padPtr->tkwin),
		PadInstanceCmd,
		(ClientData) padPtr, (void (*)(void *)) NULL);
	/*
	 * Parse the command line arguments.
	 */
	if (PadConfigure(interp, padPtr, argc-2, argv+2, 0) != TCL_OK) {
		Tk_DestroyWindow(padPtr->tkwin);
		return TCL_ERROR;
	}

	interp->result = Tk_PathName(padPtr->tkwin);
	return TCL_OK;
}

/*
 * PadInstanceCmd --
 *	This procedure implements the operations on an instance
 * 	of the widget.   It is a command procedure for the Tcl
 *	command that is the same as the name of the widget.
 *	This returns a standard Tcl result.
 */
static int
PadInstanceCmd(ClientData clientData, Tcl_Interp* interp, int argc, char** argv)
	/*ClientData clientData;	/* A pointer to a Pad struct */
	/*Tcl_Interp *interp;	/* The interpreter */
	/*int argc;		/* The number of arguments */
	/*char *argv[];		/* The command line arguments */
{
	Pad *padPtr = (Pad *)clientData;
	int result = TCL_OK;
	char c;
	int length;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}
	c = argv[1][0];
	length = strlen(argv[1]);
	if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)
			&& (length >= 2)) {
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
				argv[0], " cget option\"",
				(char *) NULL);
			return TCL_ERROR;
		}
		result = Tk_ConfigureValue(interp, padPtr->tkwin, configSpecs,
			(char *) padPtr, argv[2], 0);
	} else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)
			&& (length >= 2)) {
		if (argc == 2) {
			/*
			 * Return info about all attributes.
			 */
			result = Tk_ConfigureInfo(interp, padPtr->tkwin, configSpecs,
				(char *) padPtr, (char *) NULL, 0);
		} else if (argc == 3) {
			/*
			 * Return info about one attribute, like cget.
			 */
			result = Tk_ConfigureInfo(interp, padPtr->tkwin, configSpecs,
				(char *) padPtr, argv[2], 0);
		} else {
			/*
			 * Change one or more attributes.
			 */
			result = PadConfigure(interp, padPtr, argc-2, argv+2,
				TK_CONFIG_ARGV_ONLY);
		}
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
			"\":  must be cget, configure, position, or size",
			(char *) NULL);
		return TCL_ERROR;
	}
	return result;
}
/*
 * PadConfigure --
 *	This processes command line arguments defined in the
 *	TkConfigSpecs array.  This is called when a widget is first
 *	created or when its configuration is changed via the
 *	configure operation.
 *	It returns a standard Tcl result.
 */
static int
PadConfigure(Tcl_Interp* interp, Pad* padPtr, int argc, char** argv, int flags)
	/*Tcl_Interp *interp;	/* Needed for return values and errors */
	/*Pad *padPtr;	/* The per-instance data structure */
	/*int argc;		/* Number of valid entries in argv */
	/*char *argv[];		/* The command line arguments */
	/*int flags;		/* Tk_ConfigurePad flags */
{
	XGCValues gcValues;
	GC newGC;

	/*
	 * Tk_ConfigureWidget does all the work of parsing the
	 * arguments and looking for things in the resource database.
	 */
	if (Tk_ConfigureWidget(interp, padPtr->tkwin, configSpecs,
			argc, argv, (char *) padPtr, flags) != TCL_OK) {
		return TCL_ERROR;
	}
	/*
	 * Give the widget a default background so it doesn't get a random
	 * background between the time it is initially displayed by the
	 * X server and we get a chance to paint it.
	 */
	Tk_SetWindowBackground(padPtr->tkwin,
		Tk_3DBorderColor(padPtr->background)->pixel);
	/*
	 * Set up the graphics contexts to display the widget.  The
	 * highlight has different colors than the main area of the widget,
	 * so it requires a different context.  These contexts are all used
	 * to draw off-screen pixmaps, so turn off exposure notifications.
	 */
	#if CLOCK
	gcValues.graphics_exposures = False;
	gcValues.background = padPtr->highlight->pixel;
	newGC = Tk_GetGC(padPtr->tkwin,
		GCBackground|GCGraphicsExposures, &gcValues);
	if (padPtr->highlightGC != None) {
	    Tk_FreeGC(padPtr->display, padPtr->highlightGC);
	}
	padPtr->highlightGC = newGC;
	#endif

	gcValues.graphics_exposures = False;
	gcValues.background = Tk_3DBorderColor(padPtr->background)->pixel;
	gcValues.foreground = padPtr->foreground->pixel;
	gcValues.font = padPtr->fontPtr->fid;
	newGC = Tk_GetGC(padPtr->tkwin,
		GCBackground|GCForeground|GCFont|GCGraphicsExposures,
		&gcValues);
	if (padPtr->pixGC != None) {
	    Tk_FreeGC(padPtr->display, padPtr->pixGC);
	}
	padPtr->pixGC = newGC;

	/*
	 * Determine how big the widget wants to be and ask for that
	 * much size.  The geometry manager might gives us more or less.
	 */
	Tk_GeometryRequest(padPtr->tkwin, padPtr->width, padPtr->height);

	/*
	 * Set up a call to display ourself.
	 */
	if ((padPtr->tkwin != NULL) && Tk_IsMapped(padPtr->tkwin)
	    && !(padPtr->flags & REDRAW_PENDING)) {
		Tk_DoWhenIdle(PadDisplay, (ClientData) padPtr);
		padPtr->flags |= REDRAW_PENDING;
	}
	return TCL_OK;
}

/*
 * PadDisplay --
 *	This is called to display the pad.
 */
static void
PadDisplay(ClientData clientData)
{
	Pad *padPtr = (Pad *)clientData;
	GC gc;
	Tk_Window tkwin = padPtr->tkwin;
	#if CLOCK
	#endif

	/*
	 * Make sure the button still exists
	 * and is mapped onto the display before painting.
	 */
	padPtr->flags &= ~(REDRAW_PENDING);
	if ((padPtr->tkwin == NULL) || !Tk_IsMapped(tkwin)) {
		return;
	}

	if ( (padPtr->pixmap == NULL ) )
	{
		padPtr->width = Tk_Width(padPtr->tkwin);
		padPtr->height = Tk_Height(padPtr->tkwin);
		padPtr->pixmap = Tk_GetPixmap(padPtr->display,
			Tk_WindowId(padPtr->tkwin),
			Tk_Width(padPtr->tkwin), Tk_Height(padPtr->tkwin),
			Tk_Depth(padPtr->tkwin));
		Tk_Fill3DRectangle(padPtr->tkwin, padPtr->pixmap, padPtr->background,
			0, 0, Tk_Width(padPtr->tkwin), Tk_Height(padPtr->tkwin),
			0, TK_RELIEF_FLAT);
		//OutSource ( padPtr );
	}

	XCopyArea(padPtr->display, padPtr->pixmap, Tk_WindowId(tkwin),
	    padPtr->pixGC, 0, 0, Tk_Width(tkwin), Tk_Height(tkwin), 0, 0);
}
/*
 * PadEventProc --
 *	Handle expose and resize events.
 *	This routine is essentially the	same for all types of widgets.
 */
static void
PadEventProc(ClientData clientData, XEvent *eventPtr)
{
	Pad *padPtr = (Pad *) clientData;
	if ((eventPtr->type == Expose) && (eventPtr->xexpose.count == 0)) {
		goto redraw;
	} else if (eventPtr->type == DestroyNotify) {
		Tcl_DeleteCommand(padPtr->interp, Tk_PathName(padPtr->tkwin));
		padPtr->tkwin = NULL;
		if (padPtr->flags & REDRAW_PENDING) {
			Tk_CancelIdleCall(PadDisplay, (ClientData) padPtr);
			padPtr->flags &= ~REDRAW_PENDING;
		}
		#if CLOCK
		if (padPtr->flags & TICKING) {
			Tk_DeleteTimerHandler(padPtr->token);
			padPtr->flags &= ~TICKING;
		}
		#endif
		Tk_EventuallyFree((ClientData) padPtr, PadDestroy);
		return;
	#if CLOCK
	} else if (eventPtr->type == FocusIn) {
		if (eventPtr->xfocus.detail != NotifyPointer) {
			padPtr->flags |= GOT_FOCUS;
			if (padPtr->highlightWidth > 0) {
				goto redraw;
			}
		}
	} else if (eventPtr->type == FocusOut) {
		if (eventPtr->xfocus.detail != NotifyPointer) {
			padPtr->flags &= ~GOT_FOCUS;
			if (padPtr->highlightWidth > 0) {
				goto redraw;
			}
		}
	#endif
	} else if (eventPtr->type == ResizeRequest) {
		if ( padPtr->pixmap != NULL )
		{
			Tk_FreePixmap( padPtr->display, padPtr->pixmap );
			padPtr->pixmap = NULL;
		}
		
	}
	//return;
    
	redraw:
	if ((padPtr->tkwin != NULL) && !(padPtr->flags & REDRAW_PENDING)) {
		Tk_DoWhenIdle(PadDisplay, (ClientData) padPtr);
		padPtr->flags |= REDRAW_PENDING;
	}
}
/*
 * PadDestroy --
 *	Clean up after a pad is destroyed.
 *	This frees all the resources associated with the pad.
 */
static void
PadDestroy(char* clientData)
//PadDestroy(ClientData clientData)
    /*ClientData clientData;		/* Info about entry widget. */
{
	register Pad *padPtr = (Pad *) clientData;
    
	/*
	 * Free up all the stuff that requires special handling, then
	 * let Tk_FreeOptions handle all the standard option-related
	 * stuff.
	 */
	if (padPtr->pixGC != None) {
		Tk_FreeGC(padPtr->display, padPtr->pixGC);
	}

	if (padPtr->flags & REDRAW_PENDING) {
		Tk_CancelIdleCall(PadDisplay, (ClientData) padPtr);
	}
	/*
	 * This frees up colors and fonts and any allocated storage
	 * associated with the widget attributes.
	 */
	Tk_FreeOptions(configSpecs, (char *) padPtr, padPtr->display, 0);
	ckfree((char *) padPtr);
}

// These routines make the lower left corner (0,0) as God intended.

void PadLine ( Pad* p, int x1, int y1, int x2, int y2 )
{
	XDrawLine ( p->display, p->pixmap, p->pixGC,
		x1, p->height - y1, x2, p->height - y2 );
}

void PadRectangle ( Pad* p, int x1, int y1, int x2, int y2 )
{
	XDrawRectangle( p->display, p->pixmap, p->pixGC,
		x1, p->height - y2, x2 - x1, y2 - y1 );
}

void PadFillRectangle ( Pad* p, int x1, int y1, int x2, int y2 )
{
	XFillRectangle( p->display, p->pixmap, p->pixGC,
		x1, p->height - y2, x2 - x1, y2 - y1 );
}

void PadPolygon ( Pad* p, int n, int* v )
{
	int i;

	for ( i = 0; i < 2*n - 2; i += 2 )
		XDrawLine ( p->display, p->pixmap, p->pixGC,
			v[i], p->height - v[i+1], v[i+2], p->height - v[i+3] );

	XDrawLine ( p->display, p->pixmap, p->pixGC,
		v[2*n-2], p->height - v[2*n-1], v[0], p->height - v[1] );
}

void PadFillPolygon ( Pad* p, int n, int* v )
{
	int i;
	short w[100];

	for ( i = 0; i < 2*n; i++ )
	{
		w[i] = v[i];
		i++;
		w[i] = p->height - v[i];
	}

	XFillPolygon ( p->display, p->pixmap, p->pixGC,
		(XPoint *)w, n, Complex, CoordModeOrigin );
}

void PadCircle ( Pad* p, int x1, int y1, int radius )
{
	XDrawArc ( p->display, p->pixmap, p->pixGC,
		x1-radius, p->height - (y1+radius), 2*radius, 2*radius, 0, 360*64 );
}

void PadText ( Pad* p, int x1, int y1, char* string, int n )
{
	XDrawString ( p->display, p->pixmap, p->pixGC,
		x1, p->height - y1, string, n );
}

void PadFgColor ( Pad* p, XColor *c )
{
	p->foreground = c;
	XSetForeground ( p->display, p->pixGC, p->foreground->pixel );
}

// Need a free color?
XColor* PadGetColor( Pad* p, char* ColorName )
{
	return Tk_GetColor( p->interp, p->tkwin, Tk_GetUid ( ColorName ) );
}

void PadClear ( Pad* p )
{
	XSetForeground ( p->display, p->pixGC,
		Tk_3DBorderColor(p->background)->pixel );
	XFillRectangle ( p->display, p->pixmap, p->pixGC, 0, 0,
		p->width, p->height );
	XSetForeground ( p->display, p->pixGC, p->foreground->pixel );
}

void PadFlush ( Pad* p )
{
	XCopyArea ( p->display, p->pixmap, Tk_WindowId(p->tkwin), p->pixGC,
		0, 0, p->width, p->height, 0, 0);
	XFlush(p->display);
}

void PadLeftShift ( Pad* p, int x )
{
	if ( x > 0 )
	{
			XCopyArea ( p->display, p->pixmap, p->pixmap, p->pixGC,
				x, 0, p->width - x, p->height, 0, 0);

			// paint background color unto the area not written over
			XSetForeground ( p->display, p->pixGC,
				Tk_3DBorderColor(p->background)->pixel );
			XFillRectangle ( p->display, p->pixmap, p->pixGC,
				p->width - x, 0, x, p->height );
			XSetForeground ( p->display, p->pixGC, p->foreground->pixel );
	}
}

int PadHeight ( Pad* p )
{
	return p->height;
}

int PadWidth ( Pad* p )
{
	return p->width;
}
